rm(list=ls())

# set the seed as the last three digits of your student ID
seed = 666

if (seed == 111) {
  stop("\nThe random generator seed is still set to its default value.\nEdit the script and change it with the last three digits of your student ID\n\n", 
       call. = FALSE)
}
set.seed(seed)

# number of points in the dataset
trainSetSize = 230

################################# MAIN CODE ###########################################
cat("\n\n-------------------------------------------------------------------------------------\n")
cat("Welcome to the second PAMI homework: linear regression\n\n")

cat("For this demo we will use the \"Los Angeles ozone\" dataset. For more information\n")
cat("about it, check the following URLs:\n")
cat("(info) http://statweb.stanford.edu/~tibs/ElemStatLearn/datasets/LAozone.info.txt\n")
cat("(data) http://statweb.stanford.edu/~tibs/ElemStatLearn/datasets/LAozone.data\n\n")

cat("Let us now load the dataset and inspect its contents...")
invisible(readline(prompt = "Press [enter] to continue"))  
# uncomment the line below if you want to read the dataset from the Web
#LAozone = read.table("http://statweb.stanford.edu/~tibs/ElemStatLearn/datasets/LAozone.data",sep=",",head=T)
LAozone = read.table("LAozone.data",sep=",",head=T)
attach(LAozone)
pairs(LAozone)

cat("\n\nThe dataset contains values for the following variables in different days of the year:\n" )
cat(names(LAozone),"\n")
cat("We will now try to fit different simple univariate linear regression models\n")
cat("trying to estimate the \"ozone\" response variable from the data.\n")
invisible(readline(prompt = "Press [enter] to continue"))  

numFeats = length(names(LAozone))
for (i in 2:numFeats){
  cat("Now showing ozone ~ ", names(LAozone)[i],": ")
  fit = lm(ozone ~ LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")
  plot(LAozone[,i],ozone,xlab=names(LAozone)[i])
  abline(fit)
  invisible(readline(prompt = "Press [enter] to continue"))    
}

cat("\n\nCheck the RSS values for the different features: \"temp\" has the lowest one,\n")
cat("thus it is the best choice if we want to fit our model using only one variable.\n")
cat("Better results, though, can be obtained if we use more than one variable:\n")
cat("let us then keep \"temp\" fixed and try to add each of the other features\n")
cat("to a multivariate linear regression model.\n\n")
invisible(readline(prompt = "Press [enter] to continue"))    

# take all the columns except the one with temp
for (i in c(2:4,6:10)){ 
  cat("ozone ~ temp + ", names(LAozone)[i],": ")
  fit = lm(ozone ~ temp + LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")  
}

cat("\n\nNow, temp+ibh looks like the best choice. Let us choose a third variable...\n")
invisible(readline(prompt = "Press [enter] to continue"))    

# take all the columns except the one with temp and the one with ibh
for (i in c(2:4,7:10)){ 
  cat("ozone ~ temp + ibh + ", names(LAozone)[i],": ")
  fit = lm(ozone ~ temp + ibh + LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")  
}

cat("\n\nFor three variables, temp+ibh+humidity is the best choice. \n\n")
cat("Now let us extend this approach to all the variables: the *regsubsets*\n")
cat("command from the *leaps* library allows you to do that automatically.\n")
cat("Check out its output below here:\n\n")

invisible(readline(prompt = "Press [enter] to continue"))    
library(leaps)
regs = regsubsets(ozone ~., LAozone, method="backward")
summary(regs)

cat("\n\nAs you can see, regsubsets results are consistent with the calculations\n")
cat("we did manually. Can you tell which type of approach we used for variable\n")
cat("selection? Would the result be the same if we changed the approach in regsubset?\n\n")

cat("Let us now try to fit models up to 4 variables (rows 1-4 in the table above),\n")
cat("plus another model containing all of them.\n")

invisible(readline(prompt = "Press [enter] to continue"))    

fit1 = lm(ozone ~ temp)
fit2 = lm(ozone ~ temp + ibh)
fit3 = lm(ozone ~ temp + ibh + humidity)
fit4 = lm(ozone ~ temp + humidity + ibt + doy)
fitx = lm(ozone ~ ., data=LAozone)

cat("ozone ~ temp: MSE=", mean((predict(fit1)-ozone)^2),"\n")
cat("ozone ~ temp + ibh: MSE=", mean((predict(fit2)-ozone)^2),"\n")
cat("ozone ~ temp + ibh + humidity: MSE=", mean((predict(fit3)-ozone)^2),"\n")
cat("ozone ~ temp + humidity + ibt + doy: MSE=", mean((predict(fit4)-ozone)^2),"\n")
cat("ozone ~ ALL: MSE=", mean((predict(fitx)-ozone)^2),"\n")

cat("\nAs you can see, there is not much gain by choosing to use all the variables\n")
cat("with respect to choosing only four of them. Moreover, the model would be much\n")
cat("more complicated than needed.\n")
cat("Rather, you might want to take into account some non-linearity. For instance,\n")
cat("check the result below:\n\n")

fit5 = lm(ozone ~ temp + humidity + ibt + poly(doy,2))
cat("ozone ~ temp + humidity + ibt + poly(doy,2): MSE=", mean((predict(fit5)-ozone)^2),"\n")

cat("\n\nThis looks pretty cool: just by introducing a small nonlinear factor \n")
cat("we can get a better result than by taking all of the variables!\n")
cat("Now, the question we should ask ourselves now is: are we already overfitting?\n")
cat("To verify this, let us split the dataset in a training and a test subset, and\n")
cat("fit again our functions...\n")
invisible(readline(prompt = "Press [enter] to continue"))    
    
dataItems = nrow(LAozone)
cat("\n\nTotal number of items in the dataset: ",dataItems,"\n")
cat("Training set size: ", trainSetSize, "\n")
cat("Test set size: ", dataItems-trainSetSize, "\n\n")
train = rep(FALSE,dataItems)
train[sample(1:330, trainSetSize)] = TRUE

LA.train = LAozone[train,]
LA.test = LAozone[!train,]

fit1 = lm(ozone ~ temp, data=LA.train)
fit2 = lm(ozone ~ temp + ibh, data = LA.train)
fit3 = lm(ozone ~ temp + ibh + humidity, data = LA.train)
fit4 = lm(ozone ~ temp + humidity + ibt + doy, data = LA.train)
fitx = lm(ozone ~ ., data = LA.train)
fit5 = lm(ozone ~ temp + humidity + ibt + poly(doy,2), data = LA.train)

cat("ozone ~ temp                               : MSE.train=", mean((predict(fit1,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit1,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + ibh                         : MSE.train=", mean((predict(fit2,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit2,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + ibh + humidity              : MSE.train=", mean((predict(fit3,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit3,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + humidity + ibt + doy        : MSE.train=", mean((predict(fit4,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit4,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ ALL                                : MSE.train=", mean((predict(fitx,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fitx,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + humidity + ibt + poly(doy,2): MSE.train=", mean((predict(fit5,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit5,LA.test)-LA.test$ozone)^2), "\n")

cat("\n\nTo give you an idea about how well our function fits well the data, we now\n")
cat("plot the ozone variable as a function of temp. Black points belong to the \n")
cat("training set, green ones to test, and red ones are our predictions of the \n")
cat("test set.\n\n")
cat("You can play with the variables to see, for instance:\n")
cat("- how different training/test set sizes affect results\n")
cat("- how the plot changes if you use different functions (change the \"fit\" var)\n")
cat("- how the random initialization affects the training/test sets,\n")
cat("  and consequently the final results.")


plot(LA.train$temp,LA.train$ozone)
points(LA.test$temp,LA.test$ozone,col="green")
points(LA.test$temp,predict(fit5,LA.test),col="red")

detach(LAozone)

